
function out = anisodiff_rgb(param)
%
% Implementation based on 
%    Weickert, J.: Anisotropic diffusion in image processing (1996)
%    
%    Weickert, J.: Coherence-enhancing diffusion of colour images. Image 
%    and Vision Computing 17 (1999) 201?212
% 
%    Felsberg, M.: Autocorrelation-driven diffusion filtering. IEEE Trans. 
%     Image Processing 20 (2011) 1797:1806
%
%  Freddie Astrom, Spring 2012, freddie.astrom@liu.se
%

% ====================================================

if nargin
    k = param.K;
    iterations = param.iterations;
    dt = param.dt;
    u0 = param.u0;
    u1 = param.u1;
    verbose = param.verbose;
    opt = param.opt;
else
    
    param.dummy = 1;
    % Constants
    k = 0.01;
    iterations = 10;
    dt = 0.1;
    std = 20/256;
    r = 30;
    verbose = 1;
    opt = 'EXP';
    
    u0 = im2double(imread('8143.jpg'));
    u1 = u0;
    
    u1 = u1 + std*randn(size(u1));
    
    param.K = k;
    param.iterations = iterations;
    param.dt = dt;
    param.u0 = u0;
    param.u1 = u1;
    param.verbose = verbose;
    param.opt = opt;
    param.r = r;
end

% =================================================================
% =================================================================
%

% Initialize containers
out = param;

% =================================================================
% =================================================================
%

Ptemp = param;

% Iterate

% RUN ANISOTROPIC DIFFUSION

% gaussian kernel for regularization
uscale = 1;
h = exp( - ((-3:3)*ceil(uscale)).^2 / (2*uscale^2) );
h = h./sum(h);

g = [-1 0 1; -2 0 2; -1 0 1]/4;

hg = fspecial('gaussian',[3 3],0.5);

nExt = 10;

uk = u1;

immat(:,:,1:3,:) = u1;

for j = 1:iterations
    
    D11 = zeros(size(u1,1)+2*nExt, size(u1,2)+2*nExt);
    D12 = D11;
    D22 = D11;
    
    % Compute diffusiontensor
    for c = 1:3
        
        R = reflectImg(uk(:,:,c),nExt);
        
        Rf = conv2(extendImg(R),hg,'valid');
        
        % Rf = R; %extendImg(R);
        ux = conv2(extendImg(Rf),g,'valid');
        uy = conv2(extendImg(Rf),g','valid');
        
        % Structure tensor
        T11 = conv2(h,h,ux.^2,'same');
        T12 = conv2(h,h,ux.*uy,'same');
        T22 = conv2(h,h,uy.^2,'same');
        
        % Diffusion tensor
        z = T11-T22+1i*2*T12;
        t = T11+T22;
        l1 = (t+abs(z))./2;
        l2 = (t-abs(z))./2;
        
        c1 = exp(- l2 / k);
        c2 = exp(- l1 / k);
        
        z = sqrt(z./(abs(z)+eps));
        D11_uk = (c1 .* real(z).^2 + c2 .* imag(z).^2);
        D12_uk = -(c2-c1) .* real(z) .* imag(z);
        D22_uk = (c1 .* imag(z).^2 + c2 .* real(z).^2);
        
        D11 = D11 + D11_uk;
        D12 = D12 + D12_uk;
        D22 = D22 + D22_uk;
        
    end
    D = comp2tensor(D11/3,D12/3,D12/3,D22/3);
    
    % Update each channel
    for c = 1:3
        R = reflectImg(uk(:,:,c),nExt);
        
        uk_N = shiftmat(R,'N');
        uk_S = shiftmat(R,'S');
        uk_E = shiftmat(R,'E');
        uk_W = shiftmat(R,'W');
        
        uk_NW = shiftmat(uk_N,'W');
        uk_NE = shiftmat(uk_N,'E');
        uk_SW = shiftmat(uk_S,'W');
        uk_SE = shiftmat(uk_S,'E');
        
        M = D;
        A   = M(:,:,1);
        A_N = shiftmat(A,'N');
        A_S = shiftmat(A,'S');
        
        B   = M(:,:,2);
        B_N = shiftmat(B,'N');
        B_S = shiftmat(B,'S');
        B_E = shiftmat(B,'E');
        B_W = shiftmat(B,'W');
        
        B_NW = shiftmat(B_N,'W');
        B_NE = shiftmat(B_N,'E');
        B_SE = shiftmat(B_S,'E');
        B_SW = shiftmat(B_S,'W');
        
        C   = M(:,:,4);
        C_E = shiftmat(C,'E');
        C_W = shiftmat(C,'W');
        
        L1 =  (abs(B_SE) - B_SE)/4 + (abs(B) - B)/4;
        L2 =  (C_E + C)/2 - (abs(B_E) + abs(B))/2 ;
        L3 =  (abs(B_NE) + B_NE)/4 + (abs(B) + B)/4;
        L4 =  (A_S + A)/2 - (abs(B_S) + abs(B))/2;
        L5 = ( - (A_S + 2*A + A_N)/2 ...
            - (abs(B_SE) - B_SE + abs(B_NE) + B_NE)/4 ...
            - (abs(B_SW) + B_SW + abs(B_NW) - B_NW)/4 ...
            + (abs(B_S) + abs(B_N) + abs(B_W) + abs(B_E) + 2*abs(B))/2 ...
            - (C_W + 2*C + C_E)/2);
        L6 =  (A_N + A)/2 - (abs(B_N) + abs(B))/2;
        L7 =  (abs(B_SW) + B_SW)/4 + (abs(B) + B)/4;
        L8 =  (C_W + C)/2 - (abs(B_W) + abs(B))/2;
        L9 =  (abs(B_NW) - B_NW)/4 + (abs(B) - B)/4;
        
        update1 = L7.*uk_NW     + L4.*uk_N  + L1.*uk_NE ...
            + L8.*uk_W      + L5.*R     + L2.*uk_E  ...
            + L9.*uk_SW     + L6.*uk_S  + L3.*uk_SE;
        
        R = R + dt*update1;
        
        uk(:,:,c) = R(nExt+1:end-nExt,nExt+1:end-nExt);
    end
    
    uk(uk>1) = 1;
    uk(uk<0) = 0;
    
    immat(:,:,:,j+1) = uk;
end

if verbose
    figure(2000), clf
    subplot(221), imshow(u0), title('Original'), axis image
    subplot(222), imshow(u1), title('Noisy'), axis image
    subplot(223), imshow(uk), title('Improved'), axis image
    subplot(224), imshow(abs(u0-uk)), title('Original - Improved'), colorbar
end

out.images = immat;

end


function smat = shiftmat(mat,direction)

mat_ext = reflectImg(mat,1);

switch direction
    case 'S'
        smat = mat_ext(1:end-2,2:end-1);
    case 'N'
        smat = mat_ext(3:end,2:end-1);
    case 'W'
        smat = mat_ext(2:end-1,1:end-2);
    case 'E'
        smat = mat_ext(2:end-1,3:end);
end

end

