function out = ead(param)
%
% ead - Extended Anisotropic Diffusion
%
% Implementation scheme is based on the following publications
%   Weickert, J.: Anisotropic diffusion in image processing (1996) (p. 96)
%
%   Weickert, J.: A scheme for coherence-enhancing diffusion fltering with optimized
%   rotation invariance. Journal of Visual Communication and Image Representation
%   13 (2002) 103-118
%
% Input parameters
% param.{K,iterations,dt,verbose,u0,u1}
%   K - contrast parameter
%   dt - stepsize
%   verbose - figures
%   u0 - noisefree image
%   u1 - image with noise
%
% Freddie Astrom, Spring 2012, freddie.astrom@liu.se 
%

if nargin
    k = param.K;
    iterations = param.iterations;
    stepsize = param.dt;
    verbose = param.verbose;
    u0 = param.u0;
    u1 = param.u1;
    opt = param.opt;
    r = 20;
else
    param.dummy = 1;
    
    % Constants
    k = 0.001;
    iterations = 10;
    stepsize = 0.1;
    sgm = 10/256;
    verbose = 1;
    
    u0 = rgb2gray(im2double(imread('8143.jpg')));
    u1 = u0;
    
    u1 = u1 + sgm*randn(size(u1));
end

% gaussian kernel for regularization
uscale = 1;
h = exp( - ((-3:3)*ceil(uscale)).^2 / (2*uscale^2) );
h = h./sum(h);

g = [-1 0 1; -2 0 2; -1 0 1]/4;

hg = fspecial('gaussian',[3 3],0.5);

nExt = 1;
R = reflectImg(u1,nExt);

immat(:,:,1) = u1;

for j = 1:iterations
    
    % fprintf('Iteration %d\n',j);
    
    Rf = conv2(extendImg(R),hg,'valid');
    
    ux = conv2(extendImg(Rf),g,'valid');
    uy = conv2(extendImg(Rf),g','valid');
    
    % Structure tensor
    T11 = conv2(h,h,ux.^2,'same');
    T12 = conv2(h,h,ux.*uy,'same');
    T22 = conv2(h,h,uy.^2,'same');
    
    % Diffusion tensor
    z = T11-T22+1i*2*T12;
    t = T11+T22;
    l1 = (t+abs(z))./2;
    l2 = (t-abs(z))./2;
    
    c1 = exp(- l2 / k);
    c2 = exp(- l1 / k);
    
    z = sqrt(z./(abs(z)+eps));
    D11 = (c1 .* real(z).^2 + c2 .* imag(z).^2);
    D12 = -(c2-c1) .* real(z) .* imag(z);
    D22 = (c1 .* imag(z).^2 + c2 .* real(z).^2);
    
    D = comp2tensor(D11,D12,D12,D22);
    
    % -----------------------------------------------------
    % -----------------------------------------------------
    % -----------------------------------------------------
    % Compute structure tensor and its derivatives
    T11ux = 2*conv2(h,h,ux,'same');
    T12ux = conv2(h,h,uy,'same');
    T22ux = zeros(size(ux(:,:,1)));
    
    T11uy = zeros(size(ux));
    T12uy = conv2(h,h,ux,'same');
    T22uy = 2*conv2(h,h,uy,'same');
    
    n11 = ux.*T11ux + uy.*T12ux;
    n12 = ux.*T12ux + uy.*T22ux;
    n21 = ux.*T11uy + uy.*T12uy;
    n22 = ux.*T12uy + uy.*T22uy;
    
    N = comp2tensor(n11,n12,n21,n22);
    
    E = remapp_eigenvalues_nonsymmetric_alg(N,k);
    E(E > 1) = 1;
    E(E < 0) = 0;
    
    [e11, e12, e21, e22] = tensor2comp(E);
    
    j1 = e11.*ux + e12.*uy;
    j2 = e21.*ux + e22.*uy;
    
    gx = [-3  0 3; -10 0 10; -3  0 3]/32;
    gy = gx';
    
    update2 = conv2(extendImg(j1),gx,'valid') + conv2(extendImg(j2),gy,'valid');
    
    uk_N = shiftmat(R,'N');
    uk_S = shiftmat(R,'S');
    uk_E = shiftmat(R,'E');
    uk_W = shiftmat(R,'W');
    
    uk_NW = shiftmat(uk_N,'W');
    uk_NE = shiftmat(uk_N,'E');
    uk_SW = shiftmat(uk_S,'W');
    uk_SE = shiftmat(uk_S,'E');
    
    M = D;
    A   = M(:,:,1);
    A_N = shiftmat(A,'N');
    A_S = shiftmat(A,'S');
    
    B   = M(:,:,2);
    B_N = shiftmat(B,'N');
    B_S = shiftmat(B,'S');
    B_E = shiftmat(B,'E');
    B_W = shiftmat(B,'W');
    
    B_NW = shiftmat(B_N,'W');
    B_NE = shiftmat(B_N,'E');
    B_SE = shiftmat(B_S,'E');
    B_SW = shiftmat(B_S,'W');
    
    C   = M(:,:,4);
    C_E = shiftmat(C,'E');
    C_W = shiftmat(C,'W');
    
    L1 =  (abs(B_SE) - B_SE)/4 + (abs(B) - B)/4;
    L2 =  (C_E + C)/2 - (abs(B_E) + abs(B))/2 ;
    L3 =  (abs(B_NE) + B_NE)/4 + (abs(B) + B)/4;
    L4 =  (A_S + A)/2 - (abs(B_S) + abs(B))/2;
    L5 = ( - (A_S + 2*A + A_N)/2 ...
        - (abs(B_SE) - B_SE + abs(B_NE) + B_NE)/4 ...
        - (abs(B_SW) + B_SW + abs(B_NW) - B_NW)/4 ...
        + (abs(B_S) + abs(B_N) + abs(B_W) + abs(B_E) + 2*abs(B))/2 ...
        - (C_W + 2*C + C_E)/2);
    L6 =  (A_N + A)/2 - (abs(B_N) + abs(B))/2;
    L7 =  (abs(B_SW) + B_SW)/4 + (abs(B) + B)/4;
    L8 =  (C_W + C)/2 - (abs(B_W) + abs(B))/2;
    L9 =  (abs(B_NW) - B_NW)/4 + (abs(B) - B)/4;
    
    update1 = L7.*uk_NW     + L4.*uk_N  + L1.*uk_NE ...
        + L8.*uk_W      + L5.*R     + L2.*uk_E  ...
        + L9.*uk_SW     + L6.*uk_S  + L3.*uk_SE;
    
    
    R = R + stepsize*(2*update1 + update2);
    
    immat(:,:,j+1) = R(nExt+1:end-nExt,nExt+1:end-nExt);
    
    if verbose
        figure(2000), clf
        subplot(221), imshow(u0), title('Original'), axis image
        subplot(222), imshow(u1), title('Noisy'), axis image
        subplot(223), imshow(immat(:,:,j+1)), title('Improved'), axis image
        subplot(224), imshow(abs(u0-immat(:,:,j+1))), title('Original - Improved'), colorbar
    end
end

out = param;
out.images = immat;

end


function smat = shiftmat(mat,direction)

mat_ext = reflectImg(mat,1);

switch direction
    case 'S'
        smat = mat_ext(1:end-2,2:end-1);
    case 'N'
        smat = mat_ext(3:end,2:end-1);
    case 'W'
        smat = mat_ext(2:end-1,1:end-2);
    case 'E'
        smat = mat_ext(2:end-1,3:end);
end

end
