
function out = ead_color(param)
% 
% Implementation of proposed diffusion method in 
%
% @inproceedings{diva2:543914,
%   author = {{\AA}str{\"o}m, Freddie and Baravdish, George and Felsberg, Michael},
%   title = {{On Tensor-Based PDEs and their Corresponding Variational Formulations with Application to Color Image Denoising}},
%   booktitle = {ECCV 2012: 12th European Conference on Computer Vision, 7-12 October, Firenze, Italy},
%   year = {2012},
%   series = {Lecture Notes in Computer Science},
%   volume = {7574},
%   publisher = {Springer Berlin/Heidelberg},
% }
%
% ead_color()
%   produce example output
%  
% 
% Colormodel from
% Lenz, R., Carmona, P.L.: Hierarchical s(3)-coding of rgb histograms. In: Selected
% papers from VISAPP 2009. Volume 68. Springer (2010) 188{200
%
% Freddie Astrom, Spring 2012, freddie.astrom@liu.se
%

% ====================================================

if nargin
    K = param.K;
    iterations = param.iterations;
    dt = param.dt;
    u0 = param.u0;
    u1 = param.u1;
    verbose = param.verbose;
    opt = param.opt;
else
    
    param.dummy = 1;
    % Constants
    K = 0.01;
    K2 = 0.01;
    
    iterations = 10;
    dt = 0.05;
    std = 0.05;
    r = 30;
    verbose = 1;
    opt = 'EXP';

    u0 = im2double(imread('8143.jpg'));
    u1 = u0;
    
    u1 = u1 + std*randn(size(u1));
    u1(u1 > 1) = 1;
    u1(u1 < 0) = 0;
    
    param.K = K;
    param.K2 = K2;
    param.iterations = iterations;
    param.dt = dt;
    param.u0 = u0;
    param.u1 = u1;
    param.verbose = 0;
    param.opt = opt;
    param.r = r;
end


[rows cols dim] = size(u1);

% ====================================================
% (RGB) -> (I rho phi)

% ---------------------
% Decorrelate color space
P = 1/sqrt(6) * [sqrt(2)   sqrt(2)    sqrt(2);
                 2       -1         -1     ;
                 0       sqrt(3)    -sqrt(3)];

R = u1(:,:,1);
G = u1(:,:,2);
B = u1(:,:,3);

H = [R(:) G(:) B(:)];

F = H*P';

uk_transformed = zeros(size(u0));
uk_transformed(:,:,1) = reshape(F(:,1),[rows cols]);
uk_transformed(:,:,2) = reshape(F(:,2),[rows cols]);
uk_transformed(:,:,3) = reshape(F(:,3),[rows cols]);

% =================================================================
% =================================================================
%

% Initialize containers
out = param;

% =================================================================
% =================================================================
% 

Ptemp = param;
Kest = param.K;

% Iterate
for c = 1:3
    fprintf('COLOR: %d\n',c)
    uk = uk_transformed(:,:,c);
    Ptemp.u1 = uk;

    if c >= 2
        Ptemp.K = param.K2;
    end
    
    Ptemp_images = ead(Ptemp);
    uk_transformed_it(:,:,c,:) = Ptemp_images.images;
end

% =================================================================
% =================================================================
% (I rho phi) -> (RGB)

for k = 1:iterations+1
    % Polar coordinates for C1 and C2.
    I  = uk_transformed_it(:,:,1,k);
    C1 = uk_transformed_it(:,:,2,k);
    C2 = uk_transformed_it(:,:,3,k);
    
    S = [I(:) C1(:) C2(:)];
    
    V = S/(P');
    
    uk_rgb(:,:,1,k) = reshape(V(:,1),[rows cols]);
    uk_rgb(:,:,2,k) = reshape(V(:,2),[rows cols]);
    uk_rgb(:,:,3,k) = reshape(V(:,3),[rows cols]);
end

uk_rgb(uk_rgb>1) = 1;
uk_rgb(uk_rgb<0) = 0;

% =================================================================
% =================================================================
% Compute errors and save variables
out.images = uk_rgb;

% =================================================================
% =================================================================
% Figures
if verbose
    figure(2000), clf
    subplot(221), imagesc(u0), title('Original'), axis image
    subplot(222), imagesc(u1), title('Noisy'), axis image
    subplot(223), imagesc(uk_rgb(:,:,:,end)), title('Improved'), axis image
    subplot(224), imagesc(abs(u0-uk_rgb(:,:,:,end))), title('Original - Improved'), colorbar
end

end


