
function out = eediso_color(param)
%
% Wrapper for 
%  Michael Felsberg, "Autocorrelation-Driven Diffusion Filtering", IEEE 
%  Transactions on Image Processing, 20(7): 1797-1806, 2011.   
%
% Colormodel from
% Lenz, R., Carmona, P.L.: Hierarchical s(3)-coding of rgb histograms. In: Selected
% papers from VISAPP 2009. Volume 68. Springer (2010) 188{200
%
% Freddie Astrom, Spring 2012, freddie.astrom@liu.se
%


% ====================================================

if nargin
    K = param.K;
    iterations = param.iterations;
    dt = param.dt;
    u0 = param.u0;
    u1 = param.u1;
    verbose = param.verbose;
else
    
    param.dummy = 1;
    param.K = 0.01;
    param.K2 = 0.01;
    iterations = 100;
    
    K = param.K;
    dt = 0.05;
    std = 0.05;
    verbose = 1;
    
    u0 = im2double(imread('8143.jpg'));
    u1 = u0;
    
    u1 = u1 + std*randn(size(u1));
    
    u1(u1 > 1) = 1;
    u1(u1 < 0) = 0;
end

[rows cols dim] = size(u1);

% ====================================================
% (RGB) -> (I rho phi)

% ---------------------
% Decorrelate color space
P = 1/sqrt(6) * [sqrt(2)   sqrt(2)    sqrt(2);
    2       -1         -1     ;
    0       sqrt(3)    -sqrt(3)];

R = u1(:,:,1);
G = u1(:,:,2);
B = u1(:,:,3);

H = [R(:) G(:) B(:)];

F = H*P';

uk_transformed = zeros(size(u0));
uk_transformed(:,:,1) = reshape(F(:,1),[rows cols]);
uk_transformed(:,:,2) = reshape(F(:,2),[rows cols]);
uk_transformed(:,:,3) = reshape(F(:,3),[rows cols]);

% ====================================================
%

% Initialize containers
out = param;

% =================================================================
% =================================================================
% Autocorrelation drive diffusion
%

% Iterate
for c = 1:3
    fprintf('COLOR: %d\n',c)
    
    if c >= 2
        K = param.K2;
    end
    
    uk = uk_transformed(:,:,c);

    std = 0;
    [bld_n,psnr,s_t,kl,images] = eediso(uk, K, 1, dt, iterations, u0, std/255);
    uk_transformed_it(:,:,c,:) = images;
end

% =================================================================
% =================================================================
% (I rho phi) -> (RGB)

for k = 1:iterations+1
    % Polar coordinates for C1 and C2.
    I  = uk_transformed_it(:,:,1,k);
    C1 = uk_transformed_it(:,:,2,k);
    C2 = uk_transformed_it(:,:,3,k);
    
    S = [I(:) C1(:) C2(:)];
    
    V = S/(P');
    
    uk_rgb(:,:,1,k) = reshape(V(:,1),[rows cols]);
    uk_rgb(:,:,2,k) = reshape(V(:,2),[rows cols]);
    uk_rgb(:,:,3,k) = reshape(V(:,3),[rows cols]);
end

uk_rgb(uk_rgb>1) = 1;
uk_rgb(uk_rgb<0) = 0;


% =================================================================
% =================================================================
% Compute errors and save variables
out.images = uk_rgb;

% =================================================================
% =================================================================
% Figures
if verbose
    figure(2000), clf
    subplot(221), imagesc(u0), title('Original'), axis image
    subplot(222), imagesc(u1), title('Noisy'), axis image
    subplot(223), imagesc(uk_rgb(:,:,:,end)), title('Improved'), axis image
    subplot(224), imagesc(abs(u0-uk_rgb(:,:,:,end))), title('Original - Improved'), colorbar, axis image
end

end


