% 
% main.m
% 
% Files for reproducing results in ECCV 2012 submission
%
% @inproceedings{diva2:543914,
%   author = {{\AA}str{\"o}m, Freddie and Baravdish, George and Felsberg, Michael},
%   title = {{On Tensor-Based PDEs and their Corresponding Variational Formulations with Application to Color Image Denoising}},
%   booktitle = {ECCV 2012: 12th European Conference on Computer Vision, 7-12 October, Firenze, Italy},
%   year = {2012},
%   series = {Lecture Notes in Computer Science},
%   volume = {7574},
%   publisher = {Springer Berlin/Heidelberg},
% }
%
% 
% Implementation schemes are based on the following publications
%   Weickert, J.: Anisotropic diffusion in image processing (1996) (p. 96)
%   
%   Weickert, J.: A scheme for coherence-enhancing diffusion fltering with optimized
%     rotation invariance. Journal of Visual Communication and Image Representation
%     13 (2002) 103-118
% 
%   Felsberg, M.: Autocorrelation-driven diffusion filtering. IEEE Trans. 
%     Image Processing 20 (2011) 1797:1806
% 
% Generate tables in paper by reading the help and running:  
%   load_and_print_table.m
% 
% Generate figures in paper by reading the help and running: 
%   load_and_print_images_to_paper
% 
% 
% Freddie Astrom, Spring 2012, freddie.astrom@liu.se
%

%%

% ====================================================
% User variables

close all
clear
clc

% Add BM3D and TIP code in folder external
addpath(genpath('external/'))

%%
% Available methods
% 'CACD'    - Autocorrelation Driven Diffusion // // Obtain copy from author (see ref. in readme.txt)
% 'CEAD'    - Proposed method
% 'AD-RGB'  - Anisotropic Diffusion RGB
% 'CBM3D'   - Color version of BM3D // Obtain copy from author (see ref. in readme.txt)

iterations = 100;
run_method_vec = {'CACD' 'CEAD' 'AD-RGB' 'CBM3D'}; % 
path2save = 'results/';

% ---------------------------------------
% For each method
for m = 1:size(run_method_vec,2)
    run_method = run_method_vec(m);
   
    disp('=================================================================');
    disp('=================================================================');
    
    % ---------------------------------------
    % For each image
    %
    % Run only one index image for multiple noise levels at once
    %
    for index = [1 2 3 4] % 1 2 3 4 
        % ---------------------------------------
        % Load image
        %
        switch index
            % EVALUATION IMAGES
            case 1
                name = '8143_owl';
                % X = 250:450; Y = 50:250;
                X = 0; Y = 0;
            case 2
                name = '175043_snake_green';
                % X = 120:250; Y = 50:180;
                X = 0; Y = 0;
            case 3
                name = '208001_mushroom';
                X = 0; Y = 0;                
            case 4
                name = '87065_lizard';
                X = 0; Y = 0;
        end
        
        disp('==============================');
        
        fprintf('Image: %s\n',name);
        
        i_std = 1;
        % ---------------------------------------
        % For each standard deviation
        %
        for sgm = [5 10 20 50 70] % 5 10 20 50 70
            i_img = 1;
            i_method = 1;
                    
            [u0 u1] = load_image_color(name,X,Y,sgm);
            
            % ---------------------------------------
            % Estimate noise level from each color channel
            %
            [m,mu1R,mu2] = MST2(u1(:,:,1));
            [m,mu1G,mu2] = MST2(u1(:,:,2));
            [m,mu1B,mu2] = MST2(u1(:,:,3));
            
            sigma = (sqrt(2*mu1R) + sqrt(2*mu1G) + sqrt(2*mu1B))*255/3;  % Estimated noise std
            % sigma = sgm;                                               % True noise std
            
            Kest = (exp(1)-1)/(exp(1)-2)/255^2*sigma^2;
            
            dt = 0.2;       % Steplength
            K  = Kest/10;   % Contrast parameter in intensity channel
            K2 = Kest*10;   % Contrast parameter in color opponent channels
                            
            disp('------------------------------');
            fprintf('%s: estimated noise: %f (%d) ==> contrast parameter: %f\n',run_method{1},sigma,sgm,K);
            disp('---');
            
            % ---------------------------------------
            % Run methods
            %
            
            % Autocorrelation driven diffusion
            if any(strcmp(run_method,'CACD'))
                fprintf('## Autocorrelation driven diffusion ...  \n');
                P = struct('scheme','Tracebased','iterations',iterations,'dt',dt,'K',K,'K2',K2,'u0',u0,'u1',u1,'verbose',0);
                P = eediso_color(P);
                P.noise_est = sigma;
                P.noise_real = sgm;
                finalmat{i_std,i_img}{i_method} = P;
                i_method = i_method + 1;
                clear P
                fprintf('done \n');
            end
            
            % Proposed
            if any(strcmp(run_method,'CEAD'))
                fprintf('## Extended anisotropic diffusion ... \n');
                P = struct('scheme','Extended','iterations',iterations,'dt',dt/4,'K',K,'K2',K2,'u0',u0,'u1',u1,'verbose',0,'opt','EXP');
                P = ead_color(P);
                P.noise_est = sigma;
                P.noise_real = sgm;
                finalmat{i_std,i_img}{i_method} = P;
                i_method = i_method + 1;
                clear P
                fprintf('done \n');
            end
            
            % Non neg. scheme (RGB)
            if any(strcmp(run_method,'AD-RGB'))
                fprintf('## AD-RGB: ...  \n');
                P = struct('scheme','AD-RGB','iterations',iterations,'dt',dt/8,'K',Kest,'u0',u0,'u1',u1,'verbose',0,'opt','EXP');
                P = anisodiff_rgb(P);
                P.noise_est = sigma;
                P.noise_real = sgm;
                finalmat{i_std,i_img}{i_method} = P;
                i_method = i_method + 1;
                clear P
                fprintf('done \n');
            end
            
            % BM3D
            if any(strcmp(run_method,'CBM3D'))
                fprintf('## Block matching 3D ...  \n');
                P = struct('scheme','CBM3D','u0',u0,'u1',u1,'sigma',sigma,'verbose',0);
                [PSNR, y_est] = CBM3D(1, P.u1, P.sigma, 'np', P.verbose);
                y_est(y_est > 1) = 1;
                y_est(y_est < 0) = 0;
                P.images = y_est;
                P.noise_est = sigma;
                P.noise_real = sgm;
                finalmat{i_std,i_img}{i_method} = P;
                i_method = i_method + 1;
                clear P
                fprintf('done \n');
            end
            
            name_img{i_img} = name;
            finalmat{i_std,i_img}{i_method} = sigma; i_method = i_method + 1;
            finalmat{i_std,i_img}{i_method} = name;
            % i_img = i_img + 1;
            drawnow
            
            name_std{i_std} = sgm;
            i_std = i_std + 1;
        end

        finalmax = compute_and_find_max_values(finalmat,iterations);

        savename1 = [path2save 'finalmat_errors_estNoise_' run_method{1} '_' name ];
        save(savename1,'finalmax','-v7.3');
        clear finalmax

        savename2 = [path2save 'finalmat_estNoise_' run_method{1} '_' name ];
        save(savename2,'finalmat','-v7.3');
        clear finalmat
   end
end

